//#include <cmath> 
#include <GL/glew.h>
#include <GL/wglew.h>
#include <GL/glut.h>
#include <vector>
#include <glm/glm.hpp>
//#include <omp.h>

#pragma comment(lib, "glew32.lib")

using namespace std;  
const int width = 1024, height = 1024;

int W = 4;
int H = 4;	
float hw = W / 2.0f;
float hh = H / 2.0f;
float d = 0.2f;
int numX = (int)(W / d) + 1;    
int numY = (int)(H / d) + 1; 
const size_t total_points = (numX+1)*(numY+1);
const float timeStep = 0.05f;
int selected_index = -1;

struct Spring {
	int p1, p2;
	float rest_length;
	float Ks, Kd;
	int type;
};

vector<GLushort> indices;
vector<Spring> springs;

vector<glm::vec3> positions;
vector<glm::vec3> velocities;
vector<glm::vec3> forces;
 
vector<glm::vec3> accmForces;
vector<glm::vec3> accmVelocities;

int oldX=0, oldY=0;
float rX=15, rY=0;
int state =1 ;
float dist=-23;
const int GRID_SIZE=10;

const int STRUCTURAL_SPRING = 0;
const int SHEAR_SPRING = 1;
const int BEND_SPRING = 2;
const float DEFAULT_DAMPING = -0.125f;

int spring_count=0;
float	SstK = 4.0f,SstD = -0.5f;  //SstK = 2.5f,SstD = 1.2f;
float	SshK = 4.0f,SshD = -0.5f;
float	SflK = 2.4f,SflD = -0.8f;

glm::vec3 gravity=glm::vec3(0.0f,-0.2f,0.0f);
float mass = 5.f;

GLint viewport[4];
GLdouble MV[16];
GLdouble P[16];

glm::vec3 Up=glm::vec3(0,1,0), Right, viewDir;
float startTime =0, fps=0, avgFrames=0, maxFrames=0, minFrames=0;
int totalFrames=0;
int totalTime =0;
float totalFPS=0;

 
void AddSpring(int a, int b, float ks, float kd, int type) {
	Spring spring;
	spring.p1=a;
	spring.p2=b;
	spring.Ks=ks;
	spring.Kd=kd;
	spring.type = type;
	glm::vec3 deltaP = positions[a]-positions[b];
	spring.rest_length = sqrt(glm::dot(deltaP, deltaP));
	springs.push_back(spring);
}
void Mouse(int button, int s, int x, int y)
{
	if (s == GLUT_DOWN) 
	{
		oldX = x; 
		oldY = y; 
		int window_y = (height - y);
		float norm_y = float(window_y)/float(height/2.0);
		int window_x = x ;
		float norm_x = float(window_x)/float(width/2.0);
		
		float winZ=0;
		glReadPixels( x, height-y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ );
		if(winZ==1)
			winZ=0; 
		double objX=0, objY=0, objZ=0;
		gluUnProject(window_x,window_y, winZ,  MV,  P, viewport, &objX, &objY, &objZ);
		glm::vec3 pt(objX,objY, objZ); 
		size_t i=0;
		for(i=0;i<total_points;i++) {			 
			if( glm::distance(positions[i],pt)<0.1) {
				selected_index = i;
				printf("Intersected at %d\n",i);
				break;
			}
		}
	}	

	if(button == GLUT_MIDDLE_BUTTON)
		state = 0;
	else
		state = 1;

	if(s==GLUT_UP) {
		selected_index= -1;
		glutSetCursor(GLUT_CURSOR_INHERIT);
	}
}

void Motion(int x, int y)
{
	if(selected_index == -1) {
		if (state == 0)
			dist *= (1 + (y - oldY)/60.0f); 
		else
		{
			rY += (x - oldX)/5.0f; 
			rX += (y - oldY)/5.0f; 
		} 
	} else {
		float delta = 1500/abs(dist);
		float valX = (x - oldX)/delta; 
		float valY = (oldY - y)/delta; 
		if(abs(valX)>abs(valY))
			glutSetCursor(GLUT_CURSOR_LEFT_RIGHT);
		else 
			glutSetCursor(GLUT_CURSOR_UP_DOWN);

		positions[selected_index].x += Right[0]*valX ;
		float newValue = positions[selected_index].y+Up[1]*valY;
		if(newValue>0)
			positions[selected_index].y = newValue;
		positions[selected_index].z += Right[2]*valX + Up[2]*valY;		
	}
	oldX = x; 
	oldY = y; 

	glutPostRedisplay(); 
}


void DrawGrid()
{
	glBegin(GL_LINES);
	glColor3f(0.5f, 0.5f, 0.5f);
	for(int i=-GRID_SIZE;i<=GRID_SIZE;i++)
	{
		glVertex3f((float)i,0,(float)-GRID_SIZE);
		glVertex3f((float)i,0,(float)GRID_SIZE);

		glVertex3f((float)-GRID_SIZE,0,(float)i);
		glVertex3f((float)GRID_SIZE,0,(float)i);
	}
	glEnd();
}

void InitGL() { 
	startTime = (float)glutGet(GLUT_ELAPSED_TIME);
	glEnable(GL_DEPTH_TEST);
	int i=0, j=0, count=0;
	int l1=0, l2=0;
	float ypos = 7.0f;
	int v = numY+1;
	int u = numX+1;

	indices.resize( numX*numY*2*3);
	positions.resize(total_points);
	velocities.resize(total_points);
	forces.resize(total_points); 

	//for RK4
	accmForces.resize(total_points);
	accmVelocities.resize(total_points);

	//fill in positions
	for( j=0;j<=numY;j++) {		 
		for( i=0;i<=numX;i++) {	 
			positions[count  ].x = d*i-hw;
			positions[count  ].y = ypos;
			positions[count++].z = d*j;
		}
	}

	//fill in velocities
	memset(&(velocities[0].x),0,total_points*sizeof(glm::vec3));

	//fill in indices
	GLushort* id=&indices[0];
	for (i = 0; i < numY; i++) {        
		for (j = 0; j < numX; j++) {            
			int i0 = i * (numX+1) + j;            
			int i1 = i0 + 1;            
			int i2 = i0 + (numX+1);            
			int i3 = i2 + 1;            
			if ((j+i)%2) {                
				*id++ = i0; *id++ = i2; *id++ = i1;                
				*id++ = i1; *id++ = i2; *id++ = i3;            
			} else {                
				*id++ = i0; *id++ = i2; *id++ = i3;                
				*id++ = i0; *id++ = i3; *id++ = i1;            
			}        
		}    
	}

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	//glPolygonMode(GL_BACK, GL_LINE);
	glPointSize(5);

	wglSwapIntervalEXT(0);

	//setup springs
	// Horizontal
	for (l1 = 0; l1 < v; l1++)	// v
		for (l2 = 0; l2 < (u - 1); l2++) {
			AddSpring((l1 * u) + l2,(l1 * u) + l2 + 1,SstK,SstD,STRUCTURAL_SPRING);
		}

	// Vertical
	for (l1 = 0; l1 < (u); l1++)	
		for (l2 = 0; l2 < (v - 1); l2++) {
			AddSpring((l2 * u) + l1,((l2 + 1) * u) + l1,SstK,SstD,STRUCTURAL_SPRING);
		}

	
	// Shearing Springs
	for (l1 = 0; l1 < (v - 1); l1++)	
		for (l2 = 0; l2 < (u - 1); l2++) {
			AddSpring((l1 * u) + l2,((l1 + 1) * u) + l2 + 1,SshK,SshD,SHEAR_SPRING);
			AddSpring(((l1 + 1) * u) + l2,(l1 * u) + l2 + 1,SshK,SshD,SHEAR_SPRING);
		}

	
	// Bend Springs
	for (l1 = 0; l1 < (v); l1++) {
		for (l2 = 0; l2 < (u - 2); l2++) {
			AddSpring((l1 * u) + l2,(l1 * u) + l2 + 2,SflK,SflD,BEND_SPRING);
		}
		AddSpring((l1 * u) + (u - 3),(l1 * u) + (u - 1),SflK,SflD,BEND_SPRING);
	}
	for (l1 = 0; l1 < (u); l1++) {
		for (l2 = 0; l2 < (v - 2); l2++) {
			AddSpring((l2 * u) + l1,((l2 + 2) * u) + l1,SflK,SflD,BEND_SPRING);
		}
		AddSpring(((v - 3) * u) + l1,((v - 1) * u) + l1,SflK,SflD,BEND_SPRING);
	}
 
}

void OnReshape(int nw, int nh) {
	glViewport(0,0,nw, nh);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, (GLfloat)nw / (GLfloat)nh, 1.f, 100.0f);
	
	glGetIntegerv(GL_VIEWPORT, viewport); 
	glGetDoublev(GL_PROJECTION_MATRIX, P);

	glMatrixMode(GL_MODELVIEW);
}

void OnRender() {		
	size_t i=0;
	float current = (float) glutGet(GLUT_ELAPSED_TIME);
	++totalFrames;
	if((current-startTime)>1000)
	{		
		float elapsedTime = (current-startTime);
		fps = (totalFrames/ elapsedTime)*1000 ;
		totalFPS+=fps;
		
	    avgFrames =  (totalFPS/(++totalTime));
		char info[MAX_PATH]={0};
		sprintf_s(info, "FPS: %3.2f Min: %3.2f %3.2f %3.2f", fps, minFrames, maxFrames, avgFrames);
		glutSetWindowTitle(info); 
		startTime = current;
		totalFrames=0;
	} 
	 
	if(fps>0 && fps<minFrames )
		minFrames = fps;
	if(fps>maxFrames)
		maxFrames = fps;

	glClear(GL_COLOR_BUFFER_BIT| GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(0,0,dist);
	glRotatef(rX,1,0,0);
	glRotatef(rY,0,1,0);
	
	glGetDoublev(GL_MODELVIEW_MATRIX, MV);
	viewDir.x = (float)-MV[2];
	viewDir.y = (float)-MV[6];
	viewDir.z = (float)-MV[10];
	Right = glm::cross(viewDir, Up);

	//draw grid
	DrawGrid();
	
	//draw polygons
	glColor3f(1,1,1);
	glBegin(GL_TRIANGLES);
	for(i=0;i<indices.size();i+=3) {
		glm::vec3 p1 = positions[indices[i]];
		glm::vec3 p2 = positions[indices[i+1]];
		glm::vec3 p3 = positions[indices[i+2]];
		glVertex3f(p1.x,p1.y,p1.z);
		glVertex3f(p2.x,p2.y,p2.z);
		glVertex3f(p3.x,p3.y,p3.z);
	}
	glEnd();	 

	//draw points
	
	glBegin(GL_POINTS);
	for(i=0;i<total_points;i++) {
		glm::vec3 p = positions[i];
		int is = (i==selected_index);
		glColor3f((float)!is,(float)is,(float)is);
		glVertex3f(p.x,p.y,p.z);
	}
	glEnd();

	glutSwapBuffers();
}

void OnShutdown() {
	
}

void ComputeForces() {
	size_t i=0;
	//#pragma omp parallel for
	for(i=0;i<total_points;i++) {
		forces[i] = glm::vec3(0);
		 
		//add gravity force
		if(i!=0 && i!=( numX)	)		 
			forces[i] += gravity/mass;
		//add force due to damping of velocity
		forces[i] += DEFAULT_DAMPING*velocities[i];
	}

	//#pragma omp parallel for
	for(i=0;i<springs.size();i++) {
		glm::vec3 p1 = positions[springs[i].p1];
		glm::vec3 p2 = positions[springs[i].p2];
		glm::vec3 v1 = velocities[springs[i].p1];
		glm::vec3 v2 = velocities[springs[i].p2];
		glm::vec3 deltaP = p1-p2;
		glm::vec3 deltaV = v1-v2;
		float dist = glm::length(deltaP);
		float Hterm = (dist-springs[i].rest_length) * springs[i].Ks;
		float Dterm = (glm::dot(deltaV, deltaP) * springs[i].Kd)/dist;
		
		glm::vec3 springForce = glm::normalize(deltaP)*(Dterm - Hterm);
		 
		if(springs[i].p1 != 0 && springs[i].p1 != numX)
			forces[springs[i].p1] += springForce; 
		if(springs[i].p2 != 0 && springs[i].p2 != numX )
			forces[springs[i].p2] -= springForce;
	}
}


void IntegrateEuler(float deltaTime) {
	float deltaTimeMass = deltaTime/ mass;
	size_t i=0;

	//#pragma omp parallel for
	for(i=0;i<total_points;i++) {
		glm::vec3 oldV = velocities[i];
		velocities[i] += (forces[i]*deltaTimeMass);		
		positions[i]  += deltaTime*oldV;
		
		if(positions[i].y <0) {
			positions[i].y = 0; 
		}
	}
}
void IntegrateMidpointEuler(float deltaTime) {
	float halfDeltaT = deltaTime/2.0f;
	float deltaTimeMass = halfDeltaT/ mass;
	size_t i=0;
	for( i=0;i<total_points;i++) {
		glm::vec3 oldV = velocities[i];
		velocities[i] += (forces[i]*deltaTimeMass);		
		positions[i]  += deltaTime*oldV;
		
		if(positions[i].y <0) {
			positions[i].y = 0; 
		}
	}

	ComputeForces();

	deltaTimeMass = deltaTime/ mass;
	for(i=0;i<total_points;i++) {
		glm::vec3 oldV = velocities[i];
		velocities[i] += (forces[i]*deltaTimeMass);		
		positions[i]  += deltaTime*oldV;
		
		if(positions[i].y <0) {
			positions[i].y = 0; 
		}
	}
}

void IntegrateRK4(float deltaTime) {
	float halfDeltaT  = deltaTime/2.0f;
	float thirdDeltaT = 1/3.0f;
	float sixthDeltaT = 1/6.0f;
	float halfDeltaTimeMass = halfDeltaT/ mass;
	float deltaTimeMass = deltaTime/ mass;
	size_t i=0;

	memset(&accmForces[0],0,accmForces.size()*sizeof(glm::vec3));
	memset(&accmVelocities[0],0,accmVelocities.size()*sizeof(glm::vec3));

	for(i=0;i<total_points;i++) { 
		accmForces[i] = (forces[i]*halfDeltaTimeMass) * sixthDeltaT;		
		accmVelocities[i] = halfDeltaT*velocities[i]  * sixthDeltaT;
	}

	ComputeForces();
	
	for(i=0;i<total_points;i++) { 
		accmForces[i] = (forces[i]*halfDeltaTimeMass) * thirdDeltaT;		
		accmVelocities[i] = halfDeltaT*velocities[i]  * thirdDeltaT;		
	}

	ComputeForces();

	for(i=0;i<total_points;i++) {		
		accmForces[i] = (forces[i]*deltaTimeMass)   * thirdDeltaT;		
		accmVelocities[i] = deltaTime*velocities[i] * thirdDeltaT;		 
	}	
	
	ComputeForces();

	for(i=0;i<total_points;i++) {		 
		accmForces[i] = (forces[i]*deltaTimeMass) * sixthDeltaT;		
		accmVelocities[i] = deltaTime*velocities[i] * sixthDeltaT;
	}	

	for(i=0;i<total_points;i++) {		 
		velocities[i] += accmForces[i];		
		positions[i]  += accmVelocities[i];
		if(positions[i].y <0) {
			positions[i].y = 0; 
		}
	}			
}

void ApplyProvotDynamicInverse(float Tc=0.1f) {
	 
	//for(size_t j=0;j<10;j++) 
	for(size_t i=0;i<springs.size();i++) { 
		//check the current lengths of all springs
		glm::vec3 p1 = positions[springs[i].p1];
		glm::vec3 p2 = positions[springs[i].p2];
		glm::vec3 deltaP = p1-p2;
		
		
		float delta_length = glm::length(deltaP)-springs[i].rest_length;
		if((delta_length/springs[i].rest_length) > Tc) {
			glm::vec3 v1 = velocities[springs[i].p1];
				glm::vec3 v2 = velocities[springs[i].p2];
			if(springs[i].p1==0 || springs[i].p1 ==numX) {
				///positions[springs[i].p2] -= delta_length*0.001f*glm::normalize(v1-v2) ;				
				 
				springs[i].rest_length -= delta_length;
			} else if(springs[i].p2==0 || springs[i].p2 ==numX) {
				//positions[springs[i].p1] += delta_length*0.001f*glm::normalize(v2-v1);

				springs[i].rest_length -= delta_length;
			}else {
				
				positions[springs[i].p1] += delta_length*0.01f*glm::normalize(v2-v1)*0.5f;
				positions[springs[i].p2] -= delta_length*0.01f*glm::normalize(v2-v1)*0.5f;				
			}
		}
		/*
		//for this uncomment the outer loop
		float dist = glm::length(deltaP);
		if(dist>(springs[i].rest_length*1.01f)) {
			dist -= (springs[i].rest_length*1.01f);
			dist /= 2.0f;
			deltaP = glm::normalize(deltaP);
			deltaP *= dist;
			if(springs[i].p1==0 || springs[i].p1 ==numX) {
				positions[springs[i].p2] -= deltaP* 0.001f;
			} else if(springs[i].p2==0 || springs[i].p2 ==numX) {
				positions[springs[i].p1] += deltaP* 0.001f;
			} else {
				positions[springs[i].p1] -= deltaP*0.001f;
				positions[springs[i].p2] += deltaP*0.001f;				
			}
		}*/
	}
}
void OnIdle() {	
	ComputeForces();
	//for Explicit/Midpoint Euler 
	IntegrateEuler(timeStep);
	ApplyProvotDynamicInverse();		

	//for mid-point Euler
	//IntegrateMidpointEuler(timeStep);
	
	//for RK4
	//IntegrateRK4(timeStep);

	glutPostRedisplay();
}

 
void main(int argc, char** argv) {
	atexit(OnShutdown);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInitWindowSize(width, height);
	glutCreateWindow("GLUT Cloth Demo");

	glutDisplayFunc(OnRender);
	glutReshapeFunc(OnReshape);
	glutIdleFunc(OnIdle);
	
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);	 
	glewInit();
	InitGL();
	
	glutMainLoop();		
}
